/*
 * ${license.title}
 * Copyright (C) 2022 ${license.owner}
 * ${license.mailto}
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "rollball.hpp"
#include "fggl/gfx/camera.hpp"
#include "fggl/gfx/phong.hpp"

#include "fggl/input/camera_input.h"
#include "fggl/util/service.h"
#include "fggl/ecs3/prototype/loader.hpp"
#include "fggl/debug/draw.hpp"

#include <array>

struct Prefabs {
	fggl::ecs3::entity_t collectable;
	fggl::ecs3::entity_t player;
};

static void setup_prefabs(fggl::ecs3::World& world, Prefabs& prefabs) {

	auto storage = fggl::util::ServiceLocator::instance().get<fggl::data::Storage>();
	fggl::ecs3::load_prototype_file(world, *storage, "rollball.yml");

	{
		// player (cube because my sphere function doesn't exist yet
		prefabs.player = world.findPrototype("player");
		world.add<fggl::phys::Dynamics>(prefabs.player);

		auto* material = world.add< fggl::gfx::PhongMaterial >( prefabs.player );
		material->ambient = fggl::math::vec3(0.25F, 0.25F, 0.25F);
		material->diffuse = fggl::math::vec3(0.4F, 0.4F, 0.4F);
		material->specular = fggl::math::vec3(0.774597F, 0.774597F, 0.774597F);
		material->shininess = 0.6F;
	}

	{
		// collectable
		prefabs.collectable = world.findPrototype("collectable");
		auto* material = world.add< fggl::gfx::PhongMaterial >( prefabs.collectable );
		material->ambient = fggl::math::vec3(0.0215F, 0.1754F, 0.0215F);
		material->diffuse = fggl::math::vec3(0.007568F, 0.61424F, 0.07568F);
		material->specular = fggl::math::vec3(0.633F, 0.727811F, 0.633F);
		material->shininess = 0.6F;

		// we need both of these for callbacks to trigger.
		world.add<fggl::phys::CollisionCallbacks>(prefabs.collectable);
		world.add<fggl::phys::CollisionCache>(prefabs.collectable);
	}

}

static void setup_camera(fggl::ecs3::World& world) {
	auto prototype = world.create(false);

	// setup camera position/transform
	auto* transform = world.add<fggl::math::Transform>(prototype);
	if ( transform != nullptr) {
		transform->origin(glm::vec3(10.0f, 3.0f, 10.0f));
	}

	// setup camera components
	world.add<fggl::gfx::Camera>(prototype);
}

static fggl::ecs3::entity_t setup_environment(fggl::ecs3::World& world, fggl::math::vec2& size, demo::RollState& state) {
	{
		auto northWall = world.createFromPrototype("wallX");
		auto* transform = world.get<fggl::math::Transform>(northWall);
		transform->origin({size.x/2, 0.0f, 0.0f});
	}

	{
		auto southWall = world.createFromPrototype("wallX");
		auto* transform = world.get<fggl::math::Transform>(southWall);
		transform->origin({-size.x/2, 0.0f, 0.0f});
	}

	{
		auto westWall = world.createFromPrototype("wallZ");
		auto* transform = world.get<fggl::math::Transform>(westWall);
		transform->origin({0.0f, 0.0f, -size.y/2});
	}

	{
		auto eastWall = world.createFromPrototype("wallZ");
		auto* transform = world.get<fggl::math::Transform>(eastWall);
		transform->origin({0.0f, 0.0f, size.y/2});
	}

	{
		auto floor = world.createFromPrototype("floor");
		auto *transform = world.get<fggl::math::Transform>(floor);
		transform->origin({0.0f, -2.5f, 0.0f});
	}

	{
		// player just starts off as the prefab dictates
		state.player = world.createFromPrototype("player");
	}

	{
		// collectables
		std::array<fggl::math::vec3, 3> collectPos {{
			{-5.0f, -0.5f, 12.0f},
			{15.0f, -0.5f, 0.5f},
			{6.0f, -0.5f, -15.0f}
		}};

		// build the collectables
		int collectCount = 0;
		for (auto& pos : collectPos) {
			auto collectable = world.createFromPrototype("collectable");
			auto* transform = world.get<fggl::math::Transform>(collectable);
			transform->origin(pos);

			state.collectables[collectCount++] = collectable;
		}
	}

	return state.player;
}


namespace demo {

	RollBall::RollBall(fggl::App &app) : Game(app) {
	}

	void RollBall::activate() {
		Game::activate();

		Prefabs prefabs{};
		setup_prefabs(world(), prefabs);

		// collectable callbacks
		auto* collectableCallbacks = world().get<fggl::phys::CollisionCallbacks>(prefabs.collectable);
		collectableCallbacks->onEnter = [this](auto ourID, auto theirID) {
			if ( theirID == state.player) {
				this->world().destroy(ourID);
			}
		};

		// actual scene objects
		setup_camera(world());

		// create a 20x20 grid
		fggl::math::vec2 size{40.0f, 40.0f};
		setup_environment(world(), size, state);

	}

	void RollBall::update() {
		Game::update();
		auto& input = this->input();

			if ( state.player != fggl::ecs3::NULL_ENTITY ) {
				auto &world = this->world();

				const fggl::math::vec3 forward = fggl::math::FORWARD;
				const fggl::math::vec3 right = fggl::math::RIGHT;
				fggl::math::vec3 force = fggl::math::VEC3_ZERO;
				bool moved = false;

				if (input.keyboard.down(glfwGetKeyScancode(GLFW_KEY_W))) {
					force += forward;
					moved = true;
				}

				if (input.keyboard.down(glfwGetKeyScancode(GLFW_KEY_S))) {
					force -= forward;
					moved = true;
				}

				if (input.keyboard.down(glfwGetKeyScancode(GLFW_KEY_A))) {
					force -= right;
					moved = true;
				}

				if (input.keyboard.down(glfwGetKeyScancode(GLFW_KEY_D))) {
					force += right;
					moved = true;
				}

				if (input.keyboard.pressed(glfwGetKeyScancode(GLFW_KEY_SPACE))) {
					state.mode = (DebugMode)( ((int)state.mode + 1) % 3 );
				}

				// setup dynamic force
				if ( moved ) {
					float forceFactor = 3.0F;
					force = glm::normalize(force) * forceFactor;
					auto *dynamics = world.get<fggl::phys::Dynamics>(state.player);
					dynamics->force += force;
				}

				{
					auto cameras = world.findMatching<fggl::gfx::Camera>();
					fggl::ecs3::entity_t cam = cameras[0];
					auto *camComp = world.get<fggl::gfx::Camera>(cam);
					camComp->target = world.get<fggl::math::Transform>(state.player)->origin();

					auto *camTransform = world.get<fggl::math::Transform>(cam);
					camTransform->origin( camComp->target + cameraOffset );
				}

				// distance mode
				if ( state.mode == DebugMode::DISTANCE ) {
					closestPickup(world);
				} else {
					if ( state.closestPickup != fggl::ecs::NULL_ENTITY ) {
						if ( world.alive(state.closestPickup) ){
							auto *renderer = world.tryGet<fggl::gfx::PhongMaterial>(state.closestPickup);
							if (renderer != nullptr) {
								renderer->diffuse = {1.0f, 1.0f, 1.0f};
							}
						}

						state.closestPickup = fggl::ecs::NULL_ENTITY;
					}
				}

				spinCubes(world);
				state.time += (60.0f / 1000);
		}
	}

	void RollBall::closestPickup(fggl::ecs3::World &world) {
		float closestDistance = FLT_MAX;
		fggl::ecs::entity_t closestEntity = fggl::ecs::NULL_ENTITY;
		auto playerPos = world.get<fggl::math::Transform>(state.player)->origin();

		for ( const auto& pickup : state.collectables ) {
			if ( !world.alive(pickup) ) {
				continue;
			}

			auto* transform = world.get<fggl::math::Transform>(pickup);
			auto distance = glm::distance(transform->origin(), playerPos);
			if ( distance < closestDistance) {
				closestDistance = distance;
				closestEntity = pickup;
			}
		}

		if ( state.closestPickup != closestEntity ) {
			if ( state.closestPickup != fggl::ecs::NULL_ENTITY ){
				// deal with the previous closest pickup
				auto *renderer = world.tryGet<fggl::gfx::PhongMaterial>(state.closestPickup);
				if (renderer != nullptr) {
					renderer->diffuse = {1.0f, 1.0f, 1.0f};
				}
			}

			// now, deal with the new closest
			state.closestPickup = closestEntity;
			if (closestEntity != fggl::ecs::NULL_ENTITY) {
				auto *renderer = world.tryGet<fggl::gfx::PhongMaterial>(state.closestPickup);
				if (renderer != nullptr) {
					renderer->diffuse = {0.0f, 0.0f, 1.0f};
				}
			}
		}

		// closest pickup should face the player
		if ( state.closestPickup != fggl::ecs::NULL_ENTITY) {
			auto* transform = world.get<fggl::math::Transform>(state.closestPickup);
			auto* playerTransform = world.get<fggl::math::Transform>( state.player );
			transform->lookAt(playerTransform->origin());

			// lazy, so using the debug draw line for this bit
			dd::line( &playerTransform->origin()[0], &transform->origin()[0], dd::colors::White );
		}

	}

	void RollBall::spinCubes(fggl::ecs3::World& world) {
		// rotation
		for ( const auto& entity : state.collectables ) {
			if ( !world.alive(entity) || entity == state.closestPickup ) {
				continue;
			}

			auto* transform = world.get<fggl::math::Transform>(entity);

			// rotate the cubes
			fggl::math::vec3 angles{15, 30, 45};
			transform->rotateEuler( angles * (60.0F / 1000) );


			//auto* renderer = world.get<fggl::data::StaticMesh>( entity );
			//renderer->hintColour = fggl::math::vec3(1.0F, 1.0F, 1.0F);
		}

	}

	void RollBall::render(fggl::gfx::Graphics &gfx) {
		Game::render(gfx);
	}

} // namespace demo

